/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/** gclc:fr.bigeon.gclc.ConsoleApplication.java
 * Created on: Sep 6, 2014 */
package fr.bigeon.gclc;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.command.ICommand;
import fr.bigeon.gclc.command.ICommandProvider;
import fr.bigeon.gclc.command.SubedCommand;
import fr.bigeon.gclc.exception.CommandParsingException;
import fr.bigeon.gclc.exception.CommandRunException;
import fr.bigeon.gclc.exception.CommandRunExceptionType;
import fr.bigeon.gclc.exception.InvalidCommandName;
import fr.bigeon.gclc.i18n.Messages;
import fr.bigeon.gclc.manager.ConsoleManager;

/** <p>
 * A {@link ConsoleApplication} is an application that require the user to input
 * commands.
 * <p>
 * A typical use case is the following:
 *
 * <pre>
 * {@link ConsoleManager} manager = new {@link fr.bigeon.gclc.manager.SystemConsoleManager#SystemConsoleManager()}
 * {@link ConsoleApplication} app = new {@link ConsoleApplication#ConsoleApplication(ConsoleManager, String, String) ConsoleApplication(manager, "welcome", "see you latter")};
 * app.{@link ConsoleApplication#add(ICommand) add}("my_command", new {@link ICommand MyCommand()});
 * app.{@link ConsoleApplication#start start}();
 * </pre>
 * <p>
 * That will launch in the console application that will display "welcome",
 * return a line and prompt the user for the command until he inputs the
 * <code>exit</code> command. Then it will print "see you latter" and exit the
 * start method.
 *
 * @author Emmanuel BIGEON */
public class ConsoleApplication implements ICommandProvider {

    /** The class logger */
    private static final Logger LOGGER = Logger
            .getLogger(ConsoleApplication.class.getName());
    /** The welcome message */
    private final String header;
    /** The good bye message */
    private final String footer;
    /** The console manager */
    protected final ConsoleManager manager;
    /** The container of commands */
    private final SubedCommand root;
    /** The state of this application */
    private boolean running;
    /** The listeners */
    private final List<CommandRequestListener> listeners = new ArrayList<>();

    /** @param manager the manager
     * @param welcome the welcoming message
     * @param goodbye the goodbye message */
    public ConsoleApplication(ConsoleManager manager, String welcome,
            String goodbye) {
        this.header = welcome;
        this.footer = goodbye;
        this.manager = manager;
        root = new SubedCommand(""); //$NON-NLS-1$
    }

    @Override
    public final boolean add(ICommand cmd) throws InvalidCommandName {
        return root.add(cmd);
    }

    /** @param listener the command listener */
    public final void addListener(CommandRequestListener listener) {
        listeners.add(listener);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommandProvider#executeSub(java.lang.String,
     * java.lang.String[]) */
    @Override
    public final void executeSub(String command,
                                 String... args) throws CommandRunException {
        root.executeSub(command, args);
    }

    /** Signify to the application that no command should be inputed anymore */
    public final void exit() {
        LOGGER.fine("Request exiting application..."); //$NON-NLS-1$
        running = false;
        manager.interruptPrompt();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommandProvider#get(java.lang.String) */
    @Override
    public final ICommand get(String command) {
        return root.get(command);
    }

    /** @return the manager */
    public final ConsoleManager getManager() {
        return manager;
    }

    /** @param cmd the command
     * @throws IOException if the command could not be parsed */
    public final void interpretCommand(String cmd) throws IOException {
        List<String> args;
        try {
            args = GCLCConstants.splitCommand(cmd);
        } catch (CommandParsingException e1) {
            manager.println("Command line cannot be parsed"); //$NON-NLS-1$
            LOGGER.log(Level.INFO, "Invalid user command " + cmd, e1); //$NON-NLS-1$
            return;
        }
        if (!args.isEmpty()) {
            try {
                executeSub(args.get(0), Arrays.copyOfRange(
                        args.toArray(new String[0]), 1, args.size()));
            } catch (final CommandRunException e) {
                LOGGER.log(Level.WARNING, "Command failed: " + cmd, e); //$NON-NLS-1$
                manager.println(Messages
                        .getString("ConsoleApplication.cmd.failed", cmd)); //$NON-NLS-1$
                manager.println(e.getLocalizedMessage());
                if (e.getType() == CommandRunExceptionType.USAGE) {
                    e.getSource().help(getManager());
                }
            }
        }
    }

    /** @param listener the command listener to remove */
    public final void removeListener(CommandRequestListener listener) {
        for (int i = 0; i < listeners.size(); i++) {
            if (listeners.get(i) == listener) {
                listeners.remove(i);
                return;
            }
        }
    }

    /** Start the application */
    public final void start() {
        try {
            running = true;
            if (header != null) {
                manager.println(header);
            }
        } catch (IOException e) {
            // The manager was closed
            running = false;
            LOGGER.log(Level.WARNING,
                    "The console manager was closed. Closing the application as no one can reach it.", //$NON-NLS-1$
                    e);
            return;
        }
        do {
            runLoop();
        } while (running);
        if (footer != null) {
            try {
                manager.println(footer);
            } catch (IOException e) {
                // The manager was closed
                running = false;
                LOGGER.log(Level.WARNING,
                        "The console manager was closed.", //$NON-NLS-1$
                        e);
            }
        }
        running = false;
        LOGGER.fine("Exiting application."); //$NON-NLS-1$
    }

    /** The running loop content.
     * <p>
     * This consisting in getting the command, executing it and exiting
     * (restarting the loop). */
    private void runLoop() {
        try {
            final String cmd = manager.prompt();
            if (cmd == null || cmd.isEmpty()) {
                return;
            }
            for (final CommandRequestListener listener : listeners) {
                listener.commandRequest(cmd);
            }
            interpretCommand(cmd);
        } catch (InterruptedIOException e) {
            LOGGER.log(Level.INFO,
                    "Prompt interrupted. It is likely the application is closing.", //$NON-NLS-1$
                    e);
        } catch (IOException e) {
            // The manager was closed
            running = false;
            LOGGER.log(Level.WARNING,
                    "The console manager was closed. Closing the application as no one can reach it.", //$NON-NLS-1$
                    e);
        }
    }

    /** @return the running status */
    public boolean isRunning() {
        return running;
    }

    /** @return the root */
    public SubedCommand getRoot() {
        return root;
    }

    /** @return the header */
    public String getHeader() {
        return header;
    }

    /** @return the footer */
    public String getFooter() {
        return footer;
    }
}
